/*
 * reg.h: header for reg.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _REG_H_
#define _REG_H_

/* defined in reg.c */
extern	int	reg_exec();
extern	void	reg_comp();

#endif /* _REG_H_ */
