/*
 * hook.h: header for hook.c
 * 
 * Generated from hook.h.proto by source/Makefile
 */

#ifndef _HOOK_H_
#define _HOOK_H_

#define	ACTION_LIST $
#define	CHANNEL_NICK_LIST $
#define CHANNEL_SIGNOFF_LIST $
#define CONNECT_LIST $
#define CTCP_LIST $
#define CTCP_REPLY_LIST $
#define	DCC_CHAT_LIST $
#define	DCC_RAW_LIST $
#define DISCONNECT_LIST $
#define EXEC_LIST $
#define EXEC_ERRORS_LIST $
#define EXEC_EXIT_LIST $
#define EXEC_PROMPT_LIST $
#define FLOOD_LIST $
#define HELP_LIST $
#define	HOOK_LIST $
#define IDLE_LIST $
#define INPUT_LIST $
#define INVITE_LIST $
#define JOIN_LIST $
#define LEAVE_LIST $
#define LIST_LIST $
#define MAIL_LIST $
#define MODE_LIST $
#define MSG_LIST $
#define MSG_GROUP_LIST $
#define NAMES_LIST $
#define NICKNAME_LIST $
#define NOTE_LIST $
#define NOTICE_LIST $
#define NOTIFY_SIGNOFF_LIST $
#define NOTIFY_SIGNON_LIST $
#define PUBLIC_LIST $
#define PUBLIC_MSG_LIST $
#define PUBLIC_NOTICE_LIST $
#define PUBLIC_OTHER_LIST $
#define	RAW_IRC_LIST $
#define	SEND_ACTION_LIST $
#define	SEND_DCC_CHAT_LIST $
#define SEND_MSG_LIST $
#define SEND_NOTICE_LIST $
#define SEND_PUBLIC_LIST $
#define	SEND_TALK_LIST $
#define	SERVER_NOTICE_LIST $
#define SIGNOFF_LIST $
#define	TALK_LIST $
#define TIMER_LIST $
#define TOPIC_LIST $
#define WALL_LIST $
#define WALLOP_LIST $
#define WHO_LIST $
#define WIDELIST_LIST $
#define WINDOW_LIST $
#define WINDOW_KILL_LIST $
#define KICK_LIST $

#ifdef ON_KICK
# define NUMBER_OF_LISTS KICK_LIST + 1
#else
# define NUMBER_OF_LISTS WINDOW_KILL_LIST + 1
#endif 

extern	int	do_hook();
extern	void	save_hooks();
extern	char	*hook_info;

extern	int	in_on_who;

#endif /* _HOOK_H_ */
