/*
 * whois.h: header for whois.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _WHOIS_H_
#define _WHOIS_H_

extern	void	add_to_whois_queue();
extern	void	add_ison_to_whois();
extern	void	whois_name();
extern	void	whowas_name();
extern	void	whois_channels();
extern	void	whois_server();
extern	void	whois_oper();
extern	void	whois_lastcom();
extern	void	whois_nickname();
extern	void	whois_ignore_msgs();
extern	void	whois_ignore_notices();
extern	void	whois_ignore_walls();
extern	void	whois_ignore_invites();
extern	void	whois_join();
extern	void	whois_privmsg();
extern	void	whois_notify();
extern	void	whois_new_wallops();
extern	void	clean_whois_queue();
extern	void	set_beep_on_msg();
extern	int	beep_on_level;
extern  void    userhost_cmd_returned();
extern	void	user_is_away();
extern	void	userhost_returned();
extern	void	ison_returned();
extern	void	whois_chop();
extern	void	end_of_whois();
extern	void	whoreply();
extern	void	convert_to_whois();
extern	void	no_such_nickname();

extern	char	*redirect_format;

#define	WHOIS_WHOIS	0
#define	WHOIS_ISON	1
#define	WHOIS_USERHOST	2

#define	USERHOST_USERHOST ((void (*)()) 1)

#endif /* _WHOIS_H_ */
