/*
 * Global stuff for translation tables.
 *
 * Tomten, tomten@solace.hsh.se / tomten@lysator.liu.se
 */

#ifndef _TRANSLAT_H_
#define _TRANSLAT_H_

#if defined(__STDC__) || defined(__cplusplus)
extern	void	set_translation(char *tablename);
extern	void	enter_digraph(unsigned char key, char *str);
extern	unsigned char	get_digraph(unsigned char key);
extern	void	digraph(char *command, char *args);
extern	void	save_digraphs(FILE *fp);
#else
extern	void	set_translation();
extern	void	enter_digraph();
extern	unsigned char	get_digraph();
extern	void	digraph();
extern	void	save_digraphs();
#endif

extern	unsigned char	transToClient[256];
extern	unsigned char	transFromClient[256];
extern	char	translation;

#define DIG_TABLE_SIZE 256
extern	unsigned char	dig_table_lo[DIG_TABLE_SIZE];
extern	unsigned char	dig_table_hi[DIG_TABLE_SIZE];
extern	unsigned char	dig_table_di[DIG_TABLE_SIZE];

extern	char	digraph_hit;
extern	unsigned char	digraph_first;

#endif /* _TRANSLAT_H_ */
