/*
 * notify.h: header for notify.c
 *
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _NOTIFY_H_
#define _NOTIFY_H_

extern	void	notify();
extern	void	do_notify();
extern	void	notify_mark();
extern	void	save_notify();

#endif /* _NOTIFY_H_ */
