/*
 * exec.h: header for exec.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _EXEC_H_
#define _EXEC_H_

extern	void	check_wait_status();
extern	void	check_process_list();
extern	void	check_process_limits();
extern	void	do_processes();
extern	void	set_process_bits();
extern	int	text_to_process();
extern	char	*signals[];
extern	void	clean_up_processes();
extern	int	is_process();
extern	int	get_process_index();
extern	void	exec_server_delete();
extern	int	is_process_running();
extern	void	add_process_wait();
extern	void	set_wait_process();
extern	void	close_all_exec();
extern	int	logical_to_index();

#endif /* _EXEC_H_ */
