/*
 * vars.h: header for vars.c
 *
 * Generated from vars.h.proto automatically from the Makefile
 */

#ifndef _VARS_H_
#define _VARS_H_

extern	int	do_boolean();
extern	void	set_variable();
extern	int	get_int_var();
extern	char	*get_string_var();
extern	void	set_int_var();
extern	void	set_string_var();
extern	char	*get_string_var();
extern	void	init_variables();
extern	char	*var_settings[];
extern	char	*make_string_var();
extern	void	set_highlight_char();
extern	int	charset_size();
extern	void	save_variables();
extern	void	set_var_value();

/* var_settings indexes ... also used in display.c for highlights */
#define OFF 0
#define ON 1
#define TOGGLE 2

#define	DEBUG_COMMANDS		0x0001
#define	DEBUG_EXPANSIONS	0x0002
#define DEBUG_FUNCTIONS		0x0004

/* indexes for the irc_variable array */
#define ALWAYS_SPLIT_BIGGEST_VAR $
#define AUTO_UNMARK_AWAY_VAR $
#define AUTO_WHOWAS_VAR  $
#define BEEP_VAR $
#define BEEP_MAX_VAR $
#define BEEP_ON_MSG_VAR $
#define BEEP_WHEN_AWAY_VAR $
#define	BOLD_VIDEO_VAR $
#define CHANNEL_NAME_WIDTH_VAR $
#define CLIENTINFO_VAR $
#define CLOCK_VAR $
#define CLOCK_24HOUR_VAR $
#define CLOCK_ALARM_VAR $
#define CMDCHARS_VAR $
#define COMMAND_MODE_VAR $
#define CONTINUED_LINE_VAR $
#define DCC_BLOCK_SIZE_VAR $
#define	DEBUG_VAR $
#define DISPLAY_VAR $
#define EIGHT_BIT_CHARACTERS_VAR $
#define ENCRYPT_PROGRAM_VAR $
#define EXEC_PROTECTION_VAR $
#define FLOOD_AFTER_VAR $
#define FLOOD_RATE_VAR $
#define FLOOD_USERS_VAR $
#define FLOOD_WARNING_VAR $
#define FULL_STATUS_LINE_VAR $
#define HELP_PAGER_VAR $
#define HELP_PATH_VAR $
#define HELP_PROMPT_VAR $
#define HELP_SERVICE_VAR $
#define HELP_WINDOW_VAR $
#define HIDE_PRIVATE_CHANNELS_VAR $
#define HIGHLIGHT_CHAR_VAR $
#define HISTORY_VAR $
#define HISTORY_FILE_VAR $
#define HOLD_MODE_VAR $
#define HOLD_MODE_MAX_VAR $
#define INDENT_VAR $
#define INPUT_ALIASES_VAR $
#define INPUT_PROMPT_VAR $
#define INPUT_PROTECTION_VAR $
#define INSERT_MODE_VAR $
#define INVERSE_VIDEO_VAR $
#define LASTLOG_VAR $
#define LASTLOG_LEVEL_VAR $
#define	LOAD_PATH_VAR $
#define LOG_VAR $
#define LOGFILE_VAR $
#define MAIL_VAR $
#define MAX_RECURSIONS_VAR $
#define	MENU_VAR $
#define MINIMUM_SERVERS_VAR $
#define MINIMUM_USERS_VAR $
#define NO_CTCP_FLOOD_VAR $
#define NOTIFY_LEVEL_VAR $
#define NOTIFY_ON_TERMINATION_VAR $
#define NOVICE_VAR $
#define SCROLL_VAR $
#define SCROLL_LINES_VAR $
#define SEND_IGNORE_MSG_VAR $
#define SHELL_VAR $
#define SHELL_FLAGS_VAR $
#define SHELL_LIMIT_VAR $
#define SHOW_AWAY_ONCE_VAR $
#define SHOW_CHANNEL_NAMES_VAR $
#define SHOW_END_OF_MSGS_VAR $
#define SHOW_NUMERICS_VAR $
#define SHOW_STATUS_ALL_VAR $
#define SHOW_WHO_HOPCOUNT_VAR $
#define STATUS_AWAY_VAR $
#define STATUS_CHANNEL_VAR $
#define	STATUS_CHANOP_VAR $
#define STATUS_CLOCK_VAR $
#define STATUS_FORMAT_VAR $
#define STATUS_HOLD_VAR $
#define STATUS_HOLD_LINES_VAR $
#define STATUS_INSERT_VAR $
#define STATUS_MAIL_VAR $
#define STATUS_MODE_VAR $
#define STATUS_NOTIFY_VAR $
#define STATUS_OPER_VAR $
#define STATUS_OVERWRITE_VAR $
#define STATUS_QUERY_VAR $
#define STATUS_SERVER_VAR $
#define	STATUS_UMODE_VAR $
#define STATUS_USER_VAR $
#define STATUS_USER1_VAR $
#define STATUS_USER2_VAR $
#define STATUS_USER3_VAR $
#define STATUS_WINDOW_VAR $
#define SUPPRESS_SERVER_MOTD_VAR $
#define TAB_VAR $
#define	TAB_MAX_VAR $
/* Added for use with TERM's internal functions */
#ifdef TERM
#define TERM_COMPRESS $
#define TERM_PRIORITY $
#endif /* TERM */
/* <4/6/93> */
#define TRANSLATION_VAR $
#define UNDERLINE_VIDEO_VAR $
#define USE_OLD_MSG_VAR $
#define USER_INFO_VAR $
#define	USERINFO_VAR USER_INFO_VAR
#define	USER_WALLOPS_VAR $
#define VERBOSE_CTCP_VAR $
#define WARN_OF_IGNORES_VAR $
#define XTERM_OPTIONS_VAR $
#define NUMBER_OF_VARIABLES $

#endif /* _VARS_H_ */
