/*
 * mail.h: header for mail.c
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _MAIL_H_
#define _MAIL_H_

extern	char	*check_mail();
extern	int	check_mail_status();

#endif /* _MAIL_H_ */
