/*
 * numbers.h: header for numbers.c
 *
 * written by michael sandrof
 *
 * copyright(c) 1990 
 *
 * see the copyright file, or do a help ircii copyright 
 */

#ifndef _NUMBERS_H_
#define _NUMBERS_H_

extern	char	*numeric_banner();
extern	void	display_msg();
extern	void	numbered_command();

#endif /* _NUMBERS_H_ */
