/*
 * output.h: header for output.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _OUTPUT_H_
#define _OUTPUT_H_

extern	void	put_it();
extern	void	say();
extern	void	yell();
extern	void	help_put_it();
extern	void	send_to_server();
extern	SIGTYP	refresh_screen();
extern	void	init_screen();
extern	void	set_continued_line();
extern	FILE	*irclog_fp;
extern	void	put_file();

#endif /* _OUTPUT_H_ */
