/*
 * edit.h: header for edit.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#ifndef _EDIT_H_
#define _EDIT_H_

extern	char	*sent_nick;
extern	char	*sent_body;
extern	char	*recv_nick;
extern	void	load();
extern	void	send_text();
extern	void	eval_inputlist();
extern	void	parse_command();
extern	void	parse_line();
extern	void	edit_char();
extern	void	ExecuteTimers();

#define AWAY_ONE 0
#define AWAY_ALL 1

/* a structure for the timer list */
typedef struct	timerlist_stru
{
	int	ref;
	Long	time;
	char	*command;
	struct	timerlist_stru *next;
}	TimerList;

extern TimerList *PendingTimers;

#endif /* _EDIT_H_ */
